"""
Test Series Type.

This module contains tests for SeriesType objects.
"""

import pytest

from mokkari import exceptions
from mokkari.series import SeriesTypeList
from mokkari.session import Session


def test_series_type_list(talker: Session) -> None:
    """Test the SeriesTypeList."""
    series_types = talker.series_type_list()
    st_iter = iter(series_types)
    assert next(st_iter).name == "Annual Series"
    assert next(st_iter).name == "Cancelled Series"
    assert series_types[3].name == "Hard Cover"
    assert len(series_types) == 9


def test_bad_response_data() -> None:
    """Test for a bad series type response."""
    with pytest.raises(exceptions.ApiError):
        SeriesTypeList({"results": {"name": 1}})



"""
Variant module.

This module provides the following classes:

- Variant
- VariantSchema
"""
from marshmallow import EXCLUDE, Schema, fields, post_load


class Variant:
    """
    The Variant object contains information for variant issues.

    Args:
        **kwargs (Any): The keyword arguments is used for setting arc data from Metron.

    Attributes:
        name (int): The name of the variant cover.
        sku (str): The stock keeping unit for the variant cover.
        image (url): The url for an image for the variant cover.
    """

    def __init__(self, **kwargs):
        """Initialize a new Variant."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class VariantSchema(Schema):
    """Schema for the Variant API."""

    name = fields.Str()
    sku = fields.Str()
    image = fields.Url()

    class Meta:
        """Any unknown fields will be excluded."""

        unknown = EXCLUDE

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the Variant object.

        Args:
            data (Any): Data from Metron response.
            **kwargs (Any): Any additional keyword arguments.

        Returns:
            An :obj:`Variant` object
        """
        return Variant(**data)



"""
Test Series module.

This module contains tests for Series objects.
"""
import json
from datetime import datetime, timedelta, timezone

import pytest
import requests_mock

from mokkari import exceptions
from mokkari import series as ser
from mokkari.session import Session


def test_known_series(talker: Session) -> None:
    """Test for a known series."""
    death: ser.SeriesSchema = talker.series(1)
    assert death.name == "Death of the Inhumans"
    assert death.sort_name == "Death of the Inhumans"
    assert death.volume == 1
    assert death.year_began == 2018
    assert death.year_end == 2018
    assert death.issue_count == 5
    assert death.image == "https://static.metron.cloud/media/issue/2018/11/11/6497376-01.jpg"
    assert death.series_type.name == "Mini-Series"
    assert death.publisher.id == 1
    assert death.publisher.name == "Marvel"
    assert death.modified == datetime(
        2019,
        7,
        5,
        14,
        32,
        52,
        239629,
        tzinfo=timezone(timedelta(days=-1, seconds=72000), "-0400"),
    )
    assert death.resource_url == "https://metron.cloud/series/death-of-the-inhumans-2018/"


def test_series_without_year_end(talker: Session) -> None:
    """Test for series without a year-end date."""
    abs_carnage = talker.series(2311)
    assert abs_carnage.name == "Absolute Carnage"
    assert abs_carnage.sort_name == "Absolute Carnage"
    assert abs_carnage.volume == 1
    assert abs_carnage.year_began == 2019
    assert abs_carnage.year_end is None
    assert abs_carnage.issue_count == 5
    assert abs_carnage.publisher.id == 1
    assert abs_carnage.publisher.name == "Marvel"
    assert abs_carnage.series_type.name == "Mini-Series"


def test_serieslist(talker: Session) -> None:
    """Test the SeriesList."""
    series = talker.series_list({"name": "batman"})
    series_iter = iter(series)
    assert next(series_iter).id == 2547
    assert next(series_iter).id == 2481
    assert next(series_iter).id == 763
    assert next(series_iter).id == 93
    assert len(series) == 162
    assert series[3].id == 93
    assert series[3].display_name == "Batman (2016)"


def test_bad_series(talker: Session) -> None:
    """Test for a non-existant series."""
    with requests_mock.Mocker() as r:
        r.get(
            "https://metron.cloud/api/series/-1/",
            text='{"response_code": 404, "detail": "Not found."}',
        )
        with pytest.raises(exceptions.ApiError):
            talker.series(-1)


def test_bad_response_data() -> None:
    """Test for a bad series response."""
    with pytest.raises(exceptions.ApiError):
        ser.SeriesList({"results": {"name": 1}})


def test_bad_series_validate(talker: Session) -> None:
    """Test data with invalid data."""
    # Change the 'name' field to an int, when it should be a string.
    data = {
        "id": 150,
        "name": 50,
        "sort_name": "Gunhawks",
        "volume": 1,
        "series_type": {"id": 5, "name": "One-Shot"},
        "publisher": 1,
        "year_began": 2019,
        "year_end": 2019,
        "desc": "",
        "issue_count": 1,
        "modified": "2019-07-05T14:32:52.256872-04:00",
        "image": "https://static.metron.cloud/media/issue/2019/02/06/gunhawks-1.jpg",
    }
    with requests_mock.Mocker() as r:
        r.get(
            "https://metron.cloud/api/series/150/",
            text=json.dumps(data),
        )

        with pytest.raises(exceptions.ApiError):
            talker.series(150)


def test_series_with_associated_series(talker: Session) -> None:
    """Test series with an associated series link."""
    ff = talker.series(2818)
    assert ff.name == "Fantastic Four Annual"
    assert len(ff.associated) == 1
    assoc = ff.associated[0]
    assert assoc.id == 26
    assert assoc.name == "Fantastic Four (1961)"


def test_series_with_genres(talker: Session) -> None:
    """Test series with genres."""
    tt2011 = talker.series(3503)
    assert tt2011.name == "Teen Titans"
    assert len(tt2011.genres) == 1
    assert tt2011.genres[0].name == "Super-Hero"



"""
Creator module.

This module provides the following classes:

- Creator
- CreatorSchema
- CreatorsList
"""
from marshmallow import EXCLUDE, Schema, ValidationError, fields, post_load

from mokkari import exceptions


class Creator:
    """
    The Creator object contains information for creators.

    Args:
        **kwargs (Any): The keyword arguments is used for setting creator data from Metron.

    Attributes:
        id (int): The Metron identification number for the creator.
        name (str): The name of the creator.
        birth (date): The date of birth for the creator.
        death (date): The date of death for the creator.
        desc (str): The description of the creator.
        image (url): The url for an image associated with the creator.
        cv_id (int): Comic Vine ID for the Creator.
        resource_url (url): The url for the resource.
        modified (datetime): The date/time the creator was last changed.
    """

    def __init__(self, **kwargs):
        """Initialize a new Creator."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class CreatorSchema(Schema):
    """
    Schema for the Creator API.

    .. versionchanged:: 1.0.0

        - Added ``modified`` field

    .. versionchanged:: 2.0.2

        - Removed ``wikipedia`` field

    .. versionchanged:: 2.3.3

        - Added ``resource_url`` field.

    .. versionadded:: 2.4.0

        - Added ``cv_id`` field.
    """

    id = fields.Int()
    name = fields.Str()
    birth = fields.Date(allow_none=True)
    death = fields.Date(allow_none=True)
    desc = fields.Str()
    image = fields.Url(allow_none=True)
    cv_id = fields.Int(allow_none=True)
    resource_url = fields.URL()
    modified = fields.DateTime()

    class Meta:
        """Any unknown fields will be excluded."""

        unknown = EXCLUDE
        dateformat = "%Y-%m-%d"
        datetime = "%Y-%m-%dT%H:%M:%S%z"

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the Creator object.

        Args:
            data (Any): Data from Metron response.
            **kwargs (Any): Any additional keyword arguments.

        Returns:
            A :obj:`Creator` object.
        """
        return Creator(**data)


class CreatorsList:
    """
    The :obj:`CreatorsList` object contains a list of creators.

    Attributes:
        id (int): The Metron identification number for the creator.
        name (str): The name of the creator.
        modified (datetime): The date/time the creator was last changed.

    Returns:
        A list of creators.
    """

    def __init__(self, response) -> None:
        """Initialize a new CreatorsList."""
        self.creators = []

        schema = CreatorSchema()
        for creator_dict in response["results"]:
            try:
                result = schema.load(creator_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error) from error

            self.creators.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.creators)

    def __len__(self):
        """Return the length of the object."""
        return len(self.creators)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.creators[index]



"""
Session module.

This module provides the following classes:

- Session
"""
import platform
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urlencode

import requests
from marshmallow import ValidationError
from ratelimit import limits, sleep_and_retry
from requests.adapters import HTTPAdapter
from urllib3 import Retry

# Alias these modules to prevent namespace collision with methods.
from mokkari import __version__
from mokkari import arc as arcs
from mokkari import character as characters
from mokkari import creator as creators
from mokkari import exceptions
from mokkari import issue as issues
from mokkari import publisher as publishers
from mokkari import series as ser
from mokkari import sqlite_cache
from mokkari import team as teams

ONE_MINUTE = 60


class Session:
    """
    Session to request api endpoints.

    Args:
        username (str): The username for authentication with metron.cloud
        passwd (str): The password used for authentication with metron.cloud
        cache (SqliteCache, optional): SqliteCache to use
    """

    def __init__(
        self,
        username: str,
        passwd: str,
        cache: Optional[sqlite_cache.SqliteCache] = None,
    ) -> None:
        """Initialize a new Session."""
        self.username = username
        self.passwd = passwd
        self.header = {
            "User-Agent": f"Mokkari/{__version__} ({platform.system()}; {platform.release()})"
        }
        self.api_url = "https://metron.cloud/api/{}/"
        self.cache = cache

    def _call(
        self,
        endpoint: List[Union[str, int]],
        params: Optional[Dict[str, Union[str, int]]] = None,
    ) -> Dict[str, Any]:
        """
        Make request for api endpoints.

        Args:
            endpoint (str): The endpoint to request information from.
            params (dict[str, any]): Parameters to add to the request.
        """
        if params is None:
            params = {}

        cache_params = ""
        if params:
            ordered_params = OrderedDict(sorted(params.items(), key=lambda t: t[0]))
            cache_params = f"?{urlencode(ordered_params)}"

        url = self.api_url.format("/".join(str(e) for e in endpoint))
        cache_key = f"{url}{cache_params}"

        cached_response = self._get_results_from_cache(cache_key)
        if cached_response is not None:
            return cached_response

        data = self._request_data(url, params)

        if "detail" in data:
            raise exceptions.ApiError(data["detail"])

        self._save_results_to_cache(cache_key, data)

        return data

    def creator(self, _id: int) -> creators.Creator:
        """
        Request data for a creator based on its ``_id``.

        Args:
            _id (int): The creator id.

        Returns:
            A :obj:`Creator` object.

        Raises:
            ApiError: If there is a problem with the API request.
        """
        try:
            result = creators.CreatorSchema().load(self._call(["creator", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error) from error

        return result

    def creators_list(
        self, params: Optional[Dict[str, Union[str, int]]] = None
    ) -> creators.CreatorsList:
        """
        Request a list of creators.

        Args:
            params (dict, optional): Parameters to add to the request.

        Returns:
            A :obj:`CreatorsList` object.
        """
        res = self._get_results(["creator"], params)
        return creators.CreatorsList(res)

    def character(self, _id: int) -> characters.Character:
        """
        Request data for a character based on its ``_id``.

        Args:
            _id (int): The character id.

        Returns:
            A :obj:`Character` object.

        Raises:
            ApiError: If there is a problem with the API request.
        """
        try:
            result = characters.CharacterSchema().load(self._call(["character", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error) from error

        return result

    def characters_list(
        self, params: Optional[Dict[str, Union[str, int]]] = None
    ) -> characters.CharactersList:
        """
        Request a list of characters.

        Args:
            params (dict, optional): Parameters to add to the request.

        Returns:
            A :class:`CharactersList` object.
        """
        res = self._get_results(["character"], params)
        return characters.CharactersList(res)

    def character_issues_list(self, _id: int) -> List[issues.Issue]:
        """
        Request a list of issues that a character appears in.

        .. versionadded:: 2.2.0

        Args:
            _id (int): The character id.

        Returns:
            A list of :class:`Issue` objects.
        """
        result = self._get_results(["character", _id, "issue_list"])
        return issues.IssuesList(result)

    def publisher(self, _id: int) -> publishers.Publisher:
        """
        Request data for a publisher based on its ``_id``.

        Args:
            _id (int): The publisher id.

        Returns:
            A :obj:`Publisher` object.

        Raises:
            ApiError: If there is a problem with the API request.
        """
        try:
            result = publishers.PublisherSchema().load(self._call(["publisher", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error) from error

        return result

    def publishers_list(
        self, params: Optional[Dict[str, Union[str, int]]] = None
    ) -> publishers.PublishersList:
        """
        Request a list of publishers.

        Args:
            params (dict, optional): Parameters to add to the request.

        Returns:
            A :class:`PublishersList` object.
        """
        res = self._get_results(["publisher"], params)
        return publishers.PublishersList(res)

    def team(self, _id: int) -> teams.Team:
        """
        Request data for a team based on its ``_id``.

        Args:
            _id (int): The team id.

        Returns:
            A :obj:`Team` object.

        Raises:
            ApiError: If there is a problem with the API request.
        """
        try:
            result = teams.TeamSchema().load(self._call(["team", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error) from error

        return result

    def teams_list(
        self, params: Optional[Dict[str, Union[str, int]]] = None
    ) -> teams.TeamsList:
        """
        Request a list of teams.

        Args:
            params (dict, optional): Parameters to add to the request.

        Returns:
            A :class:`TeamsList` object.
        """
        res = self._get_results(["team"], params)
        return teams.TeamsList(res)

    def team_issues_list(self, _id: int) -> List[issues.Issue]:
        """
        Request a list of issues that a team appears in.

        .. versionadded:: 2.2.0

        Args:
            _id (int): The team id.

        Returns:
            A list of :class:`Issue` objects.
        """
        result = self._get_results(["team", _id, "issue_list"])
        return issues.IssuesList(result)

    def arc(self, _id: int) -> arcs.Arc:
        """
        Request data for a story arc based on its ``_id``.

        Args:
            _id (int): The story arc id.

        Returns:
            A :obj:`Arc` object.

        Raises:
            ApiError: If there is a problem with the API request.
        """
        try:
            result = arcs.ArcSchema().load(self._call(["arc", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error) from error

        return result

    def arcs_list(self, params: Optional[Dict[str, Union[str, int]]] = None) -> arcs.ArcsList:
        """
        Request a list of story arcs.

        Args:
            params (dict, optional): Parameters to add to the request.

        Returns:
            A :class:`ArcsList` object.
        """
        res = self._get_results(["arc"], params)
        return arcs.ArcsList(res)

    def arc_issues_list(self, _id: int) -> List[issues.Issue]:
        """
        Request a list of issues for a story arc.

        Args:
            _id (int): The arc id.

        Returns:
            A list of :class:`Issue` objects.
        """
        result = self._get_results(["arc", _id, "issue_list"])
        return issues.IssuesList(result)

    def series(self, _id: int) -> ser.Series:
        """
        Request data for a series based on its ``_id``.

        Args:
            _id (int): The series id.

        Returns:
            A :obj:`Series` object.

        Raises:
            ApiError: If there is a problem with the API request.
        """
        try:
            result = ser.SeriesSchema().load(self._call(["series", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error) from error

        return result

    def series_list(
        self, params: Optional[Dict[str, Union[str, int]]] = None
    ) -> ser.SeriesList:
        """
        Request a list of series.

        Args:
            params (dict, optional): Parameters to add to the request.

        Returns:
            A :class:`SeriesList` object.
        """
        res = self._get_results(["series"], params)
        return ser.SeriesList(res)

    def series_type_list(
        self, params: Optional[Dict[str, Union[str, int]]] = None
    ) -> ser.SeriesTypeList:
        """
        Request a list of series types.

        .. versionadded:: 2.2.2

            - Add ``series_type_list`` method

        Args:
            params (dict, optional): Parameters to add to the request.

        Returns:
            A :class:`SeriesTypeList` object.
        """
        res = self._get_results(["series_type"], params)
        return ser.SeriesTypeList(res)

    def issue(self, _id: int) -> issues.Issue:
        """
        Request data for an issue based on it's ``_id``.

        Args:
            _id (int): The issue id.

        Returns:
            A :obj:`Issue` object.

        Raises:
            ApiError: If there is a problem with the API request.
        """
        try:
            result = issues.IssueSchema().load(self._call(["issue", _id]))
        except ValidationError as error:
            raise exceptions.ApiError(error) from error

        return result

    def issues_list(
        self, params: Optional[Dict[str, Union[str, int]]] = None
    ) -> issues.IssuesList:
        """
        Request a list of issues.

        Args:
            params (dict, optional): Parameters to add to the request.

        Returns:
            A :class:`IssuesList` object.
        """
        res = self._get_results(["issue"], params)
        return issues.IssuesList(res)

    def role_list(
        self, params: Optional[Dict[str, Union[str, int]]] = None
    ) -> issues.RoleList:
        """
        Request a list of creator roles.

        Args:
            params (dict, optional): Parameters to add to the request.

        Returns:
            A :class:`RoleList` object.

        """
        res = self._get_results(["role"], params)
        return issues.RoleList(res)

    def _get_results(
        self,
        endpoint: List[Union[str, int]],
        params: Optional[Dict[str, Union[str, int]]] = None,
    ) -> Dict[str, Any]:
        if params is None:
            params = {}

        result = self._call(endpoint, params=params)
        if result["next"]:
            result = self._retrieve_all_results(result)
        return result

    def _retrieve_all_results(self, data: Dict[str, Any]) -> Dict[str, Any]:
        has_next_page = True
        next_page = data["next"]

        while has_next_page:
            if cached_response := self._get_results_from_cache(next_page):
                data["results"].extend(cached_response["results"])
                if cached_response["next"]:
                    next_page = cached_response["next"]
                else:
                    has_next_page = False
                continue

            response = self._request_data(next_page)
            data["results"].extend(response["results"])

            self._save_results_to_cache(next_page, response)

            if response["next"]:
                next_page = response["next"]
            else:
                has_next_page = False

        return data

    @sleep_and_retry
    @limits(calls=25, period=ONE_MINUTE)
    def _request_data(
        self, url: str, params: Optional[Dict[str, Union[str, int]]] = None
    ) -> Any:
        if params is None:
            params = {}

        try:
            session = requests.Session()
            retry = Retry(connect=3, backoff_factor=0.5)
            session.mount("https://", HTTPAdapter(max_retries=retry))
            response = session.get(
                url,
                params=params,
                timeout=2.5,
                auth=(self.username, self.passwd),
                headers=self.header,
            ).json()
        except requests.exceptions.ConnectionError as e:
            raise exceptions.ApiError(f"Connection error: {repr(e)}") from e

        return response

    def _get_results_from_cache(self, key: str) -> Optional[Any]:
        cached_response = None

        if self.cache:
            try:
                cached_response = self.cache.get(key)
            except AttributeError as e:
                raise exceptions.CacheError(
                    f"Cache object passed in is missing attribute: {repr(e)}"
                ) from e

        return cached_response

    def _save_results_to_cache(self, key: str, data: str) -> None:
        if self.cache:
            try:
                self.cache.store(key, data)
            except AttributeError as e:
                raise exceptions.CacheError(
                    f"Cache object passed in is missing attribute: {repr(e)}"
                ) from e



"""
Test Cache module.

This module contains tests for SqliteCache objects.
"""
import pytest
import requests_mock

from mokkari import api, exceptions


class NoGet:
    """The NoGet object fakes storing data from the sqlite cache."""

    def store(self, key, value) -> None:
        """Save no data."""
        # This method should store key value pair
        return


class NoStore:
    """The NoStore object fakes getting data from the sqlite cache."""

    def get(self, key) -> None:
        """Retrieve no data."""
        return None


def test_no_get(dummy_username: str, dummy_password: str) -> None:
    """Test for retrieving failure."""
    m = api(username=dummy_username, passwd=dummy_password, cache=NoGet())

    with pytest.raises(exceptions.CacheError):
        m.series(5)


def test_no_store(dummy_username: str, dummy_password: str) -> None:
    """Test for saving data error."""
    m = api(username=dummy_username, passwd=dummy_password, cache=NoStore())

    with requests_mock.Mocker() as r:
        r.get(
            "https://metron.cloud/api/series/5/",
            text='{"response_code": 200}',
        )

        with pytest.raises(exceptions.CacheError):
            m.series(5)


# def test_sql_store(dummy_username: str, dummy_password: str) -> None:
#     """Test for saving data to the sqlite cache."""
#     fresh_cache = sqlite_cache.SqliteCache(":memory:")
#     test_cache = sqlite_cache.SqliteCache("tests/testing_mock.sqlite")

#     m = api(username=dummy_username, passwd=dummy_password, cache=fresh_cache)
#     url = "https://metron.cloud/api/series/1/"

#     assert fresh_cache.get(url) is None

#     try:
#         with requests_mock.Mocker() as r:
#             r.get(url, text=json.dumps(test_cache.get(url)))
#             m.series(1)

#         assert fresh_cache.get(url) is not None
#     except TypeError:
#         print(
#             "This test will fail after cache db deleted.\n"
#             "It should pass if you now re-run the test suite without deleting the database."
#         )
#         assert False



"""
Test Characters module.

This module contains tests for Character objects.
"""
import json
from datetime import date, datetime, timedelta, timezone

import pytest
import requests_mock

from mokkari import character, exceptions
from mokkari.session import Session


def test_known_character(talker: Session) -> None:
    """Test for a known character."""
    black_bolt: character.CharacterSchema = talker.character(1)
    assert black_bolt.name == "Black Bolt"
    assert (
        black_bolt.image
        == "https://static.metron.cloud/media/character/2018/11/11/black-bolt.jpg"
    )
    assert len(black_bolt.creators) == 2
    assert len(black_bolt.teams) == 3
    assert black_bolt.modified == datetime(
        2021,
        9,
        9,
        15,
        52,
        49,
        90281,
        tzinfo=timezone(timedelta(days=-1, seconds=72000), "-0400"),
    )
    assert black_bolt.resource_url == "https://metron.cloud/character/black-bolt/"


def test_characterlist(talker: Session) -> None:
    """Test the CharactersList."""
    character = talker.characters_list({"name": "man"})
    character_iter = iter(character)
    assert next(character_iter).name == "'Mazing Man"
    assert next(character_iter).name == "3-D Man (Chandler)"
    assert next(character_iter).name == "3-D Man (Garrett)"
    assert len(character) == 576
    assert character[2].name == "3-D Man (Garrett)"


def test_character_issue_list(talker: Session) -> None:
    """Test for getting an issue list for an arc."""
    issues = talker.character_issues_list(1)
    assert len(issues) == 344
    assert issues[0].id == 258
    assert issues[0].issue_name == "Fantastic Four (1961) #45"
    assert issues[0].cover_date == date(1965, 12, 1)


def test_bad_character(talker: Session) -> None:
    """Test for a non-existing character."""
    with requests_mock.Mocker() as r:
        r.get(
            "https://metron.cloud/api/character/-1/",
            text='{"response_code": 404, "detail": "Not found."}',
        )
        with pytest.raises(exceptions.ApiError):
            talker.character(-1)


def test_bad_response_data() -> None:
    """Test for a bad character response."""
    with pytest.raises(exceptions.ApiError):
        character.CharactersList({"results": {"name": 1}})


def test_bad_character_validate(talker: Session) -> None:
    """Test data with invalid data."""
    # Change the 'name' field to an int, when it should be a string.
    data = {
        "id": 150,
        "name": 50,
        "alias": [],
        "desc": "Foo",
        "image": "https://static.metron.cloud/media/character/2018/11/15/moon-knight.jpg",
        "creators": [
            {"id": 146, "name": "Doug Moench", "modified": "2019-06-23T15:13:21.994966-04:00"}
        ],
        "teams": [],
        "modified": "2020-07-29T17:48:36.347982-04:00",
    }

    with requests_mock.Mocker() as r:
        r.get(
            "https://metron.cloud/api/character/150/",
            text=json.dumps(data),
        )

        with pytest.raises(exceptions.ApiError):
            talker.character(150)



"""
Series module.

This module provides the following classes:

- SeriesType
- SeriesTypeSchema
- Series
- SeriesSchema
- SeriesList
"""
from marshmallow import EXCLUDE, Schema, ValidationError, fields, post_load

from mokkari import exceptions
from mokkari.genre import GenreSchema
from mokkari.publisher import PublisherSchema


class SeriesType:
    """
    The SeriesType object contains information for type of series.

    Args:
        **kwargs (Any): The keyword arguments is used for setting series type data from Metron.

    Attributes:
        id (int): The Metron identification number for the series type.
        name (str): The name of the series type.
    """

    def __init__(self, **kwargs):
        """Initialize a new SeriesType."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class SeriesTypeSchema(Schema):
    """Schema for the Series Type."""

    id = fields.Int()
    name = fields.Str()

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the SeriesType object.

        Args:
            data (Any): Data from Metron response.
            **kwargs (Any): Any additional keyword arguments.

        Returns:
            A :obj:`SeriesType` object.
        """
        return SeriesType(**data)


class SeriesTypeList:
    """
    The :obj:`SeriesTypeList` object contains a list of series types.

    Attributes:
        id (int): The Metron identification number for the series type.
        name (str): The name of the series type.

    Returns:
        A list of series types.
    """

    def __init__(self, response):
        """Initialize a new SeriesTypeList."""
        self.series = []

        schema = SeriesTypeSchema()
        for series_type_dict in response["results"]:
            try:
                result = schema.load(series_type_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error) from error

            self.series.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.series)

    def __len__(self):
        """Return the length of the object."""
        return len(self.series)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.series[index]


class AssociatedSeries:
    """
    The AssociateSeries objects contains any associated series to the primary series.

    Args:
        **kwargs (Any): The keyword arguments is used for setting associated series data
        from Metron.

    Attributes:
        id (int): The Metron identification number for the associated series.
        name (str): The name of the associated series.
    """

    def __init__(self, **kwargs):
        """Initialize a new AssociatedSeries."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class AssociatedSeriesSchema(Schema):
    """Schema for Associated Series."""

    id = fields.Int()
    name = fields.Str(data_key="series")

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the AssociatedSeries object.

        Args:
            data (Any): Data from Metron response.
            **kwargs (Any): Any additional keyword arguments.

        Returns:
            An :obj:`AssociatedSeries` object.
        """
        return AssociatedSeries(**data)


class Series:
    """
    The Series object contains information for comic series.

    Args:
        **kwargs (Any): The keyword arguments is used for setting series data from Metron.

    Attributes:
        id (int): The Metron identification number for the series.
        name (str): The name of the series.
        sort_name (str): The name used to sort a series.
        volume (int): The volume number for a series.
        publisher (Publisher): The publisher of the series.
        year_began (int): The cover year the series began.
        year_end (int, optional): The cover year in which the series ended.
        desc (str): A summary description of the series.
        issue_count (int): The number of issues the series contains.
        display_name (str): The display name for the series.
        genres (list[Genre]): A list of genres for the series.
        associated (list[AssociatedSeries]): A list of of series associated with the
        primary series.
        cv_id (int): Comic Vine ID for the series.
        resource_url (url): The url for the resource.
        modified (datetime): The date/time the series was last changed.
    """

    def __init__(self, **kwargs):
        """Initialize a new Series."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class SeriesSchema(Schema):
    """
    Schema for the Series API.

    .. versionchanged:: 1.0.0

        - Added ``modified`` field

    .. versionchanged:: 1.0.5

        - Added ``associated`` field

    .. versionchanged:: 2.0.0

        - Changed ``publisher`` to a nested field.

    .. versionchanged:: 2.0.3
        - Changed ``series_type`` to a string field.

    .. versionchanged:: 2.0.4
        - Reverted ``series_type`` back to a nested field.

    .. versionchanged:: 2.1.1
        Added ``genres`` fields

    .. versionchanged:: 2.3.3

        - Added ``resource_url`` field.

    .. versionadded:: 2.4.0

        - Added ``cv_id`` field.
    """

    id = fields.Int()
    name = fields.Str()
    sort_name = fields.Str()
    volume = fields.Int()
    series_type = fields.Nested(SeriesTypeSchema)
    publisher = fields.Nested(PublisherSchema)
    year_began = fields.Int()
    year_end = fields.Int(allow_none=True)
    desc = fields.Str()
    issue_count = fields.Int()
    image = fields.Url(allow_none=True)
    display_name = fields.Str(data_key="series")
    genres = fields.Nested(GenreSchema, many=True)
    associated = fields.Nested(AssociatedSeriesSchema, many=True)
    cv_id = fields.Int(allow_none=True)
    resource_url = fields.URL()
    modified = fields.DateTime()

    class Meta:
        """Any unknown fields will be excluded."""

        unknown = EXCLUDE
        datetime = "%Y-%m-%dT%H:%M:%S%z"

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the Series object.

        Args:
            data (Any): Data from Metron response.
            **kwargs (Any): Any additional keyword arguments.

        Returns:
            A :obj:`Series` object.
        """
        return Series(**data)


class SeriesList:
    """
    The :obj:`SeriesList` object contains a list of series.

    Attributes:
        id (int): The Metron identification number for the series.
        series (str): The name of the series.
        modified (datetime): The date/time the series was last changed.

    Returns:
        A list of series.
    """

    def __init__(self, response):
        """Initialize a new SeriesList."""
        self.series = []

        schema = SeriesSchema()
        for series_dict in response["results"]:
            try:
                result = schema.load(series_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error) from error

            self.series.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.series)

    def __len__(self):
        """Return the length of the object."""
        return len(self.series)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.series[index]



"""
Genre Module.

This module provides the following classes:

- Genre
- GenreSchema
"""
from marshmallow import EXCLUDE, Schema, fields, post_load


class Genre:
    """
    The Genre object contains information about a Series or Issue genre.

    Args:
        **kwargs (Any): The keyword arguments is used for setting genre data from Metron.

    Attributes:
        id (int): The Metron identification number for the genre.
        name (str): The name of the genre.
    """

    def __init__(self, **kwargs):
        """Initialize a new Genre."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class GenreSchema(Schema):
    """Schema for the Genre."""

    id = fields.Int()
    name = fields.Str()

    class Meta:
        """Any unknown fields will be excluded."""

        unknown = EXCLUDE

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the Genre object.

        Args:
            data (Any): Data from Metron response.
            **kwargs (Any): Any additional keyword arguments.

        Returns:
            A :obj:`Genre` object.
        """
        return Genre(**data)



"""Project entry file."""

# Keep this at beginning of file to prevent circular import with session
__version__ = "2.4.0"

from typing import Optional

from mokkari import exceptions, session, sqlite_cache


def api(
    username: Optional[str] = None,
    passwd: Optional[str] = None,
    cache: sqlite_cache.SqliteCache = None,
) -> session.Session:
    """Entry function the sets login credentials for metron.cloud.

    Args:
        username (str): The username used for metron.cloud.
        passwd (str): The password used for metron.cloud.
        SqliteCache optional: SqliteCache to use

    Returns:
        A :obj:`Session` object

    Raises:
        AuthenticationError: If Metron returns with an invalid API credentials response.
    """
    if username is None:
        raise exceptions.AuthenticationError("Missing username.")

    if passwd is None:
        raise exceptions.AuthenticationError("Missing passwd.")

    return session.Session(username, passwd, cache=cache)



"""
Arc module.

This module provides the following classes:

- Arc
- ArcSchema
- ArcsList
"""
from marshmallow import EXCLUDE, Schema, ValidationError, fields, post_load

from mokkari import exceptions


class Arc:
    """
    The Arc object contains information for story arcs.

    Args:
        **kwargs (Any): The keyword arguments is used for setting arc data from Metron.

    Attributes:
        id (int): The Metron identification number for the story arc.
        name (str): The name of the story arc.
        desc (str): The description of the story arc.
        image (url): The url for an image associated with the story arc.
        cv_id (int): Comic Vine ID for the story arc.
        resource_url (url): The url for the resource.
        modified (datetime): The date/time the story arc was last changed.
    """

    def __init__(self, **kwargs):
        """Initialize a new Arc."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class ArcSchema(Schema):
    """
    Schema for the Arc API.

    .. versionchanged:: 1.0.0

        - Added ``modified`` field

    .. versionchanged:: 2.3.3

        - Added ``resource_url`` field.

    .. versionadded:: 2.4.0

        - Added ``cv_id`` field.
    """

    id = fields.Int()
    name = fields.Str()
    desc = fields.Str()
    image = fields.Url(allow_none=True)
    cv_id = fields.Int(allow_none=True)
    resource_url = fields.URL()
    modified = fields.DateTime()

    class Meta:
        """Any unknown fields will be excluded."""

        unknown = EXCLUDE
        datetime = "%Y-%m-%dT%H:%M:%S%z"

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the arc object.

        Args:
            data (Any): Data from Metron response.
            **kwargs (Any): Any additional keyword arguments.

        Returns:
            An :obj:`Arc` object
        """
        return Arc(**data)


class ArcsList:
    """
    The :obj:`ArcsList` object contains a list of story arcs.

    Attributes:
        id (int): The Metron identification number for the story arc.
        name (str): The name of the story arc.
        modified (datetime): The date/time the story arc was last changed.

    Returns:
        A list of story arcs.
    """

    def __init__(self, response):
        """Initialize a new ArcsList."""
        self.arcs = []

        schema = ArcSchema()
        for arc_dict in response["results"]:
            try:
                result = schema.load(arc_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error) from error

            self.arcs.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.arcs)

    def __len__(self):
        """Return the length of the object."""
        return len(self.arcs)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.arcs[index]



"""
SQLite Cache module.

This module provides the following classes:

- SqliteCache
"""
import json
import sqlite3
from datetime import datetime, timedelta
from typing import Any, Optional


class SqliteCache:
    """
    The SqliteCache object to cache search results from Metron.

    Args:
        db_name (str): Path and database name to use.
        expire (int, optional): The number of days to keep the cache results
        before they expire.
    """

    def __init__(
        self, db_name: str = "mokkari_cache.db", expire: Optional[int] = None
    ) -> None:
        """Initialize a new SqliteCache."""
        self.expire = expire
        self.con = sqlite3.connect(db_name)
        self.cur = self.con.cursor()
        self.cur.execute("CREATE TABLE IF NOT EXISTS responses (key, json, expire)")
        self.cleanup()

    def get(self, key: str) -> Optional[Any]:
        """
        Retrieve data from the cache database.

        Args:
            key (str): value to search for.
        """
        self.cur.execute("SELECT json FROM responses WHERE key = ?", (key,))
        return json.loads(result[0]) if (result := self.cur.fetchone()) else None

    def store(self, key: str, value: str) -> None:
        """
        Save data to the cache database.

        Args:
            key (str): Item id.
            value (str): data to save.
        """
        self.cur.execute(
            "INSERT INTO responses(key, json, expire) VALUES(?, ?, ?)",
            (key, json.dumps(value), self._determine_expire_str()),
        )
        self.con.commit()

    def cleanup(self) -> None:
        """Remove any expired data from the cache database."""
        if not self.expire:
            return
        self.cur.execute(
            "DELETE FROM responses WHERE expire < ?;",
            (datetime.now().strftime("%Y-%m-%d"),),
        )
        self.con.commit()

    def _determine_expire_str(self) -> str:
        if self.expire:
            dt = datetime.now() + timedelta(days=self.expire)
        else:
            dt = datetime.now()
        return dt.strftime("%Y-%m-%d")



"""
Reprint module.

This module provides the following classes:

- Reprint
- ReprintSchema
- ReprintsList
"""
from marshmallow import EXCLUDE, Schema, fields, post_load


class Reprint:
    """
    The Reprint object contains information for a reprint issue.

    Args:
        **kwargs (Any): The keyword arguments is used for setting reprint issue
        data from Metron.

    Attributes:
        id (int): The Metron identification number for the issue.
        issue (str): The name of the issue being reprinted.
    """

    def __init__(self, **kwargs):
        """Initialize a new Reprint."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class ReprintSchema(Schema):
    """Schema for the Reprint API."""

    id = fields.Int()
    issue = fields.Str()

    class Meta:
        """Any unknown fields will be excluded."""

        unknown = EXCLUDE

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the reprint object.

        Args:
            data (Any): Data from Metron response.
            **kwargs (Any): Any additional keyword arguments.

        Returns:
            A :obj:`Reprint` object.
        """
        return Reprint(**data)



"""
Rating module.

This module provides the following classes:

- Rating
- RatingSchema
"""
from marshmallow import EXCLUDE, Schema, fields, post_load


class Rating:
    """
    The rating object contains information for issue's rating.

    Args:
        **kwargs (Any): The keyword arguments is used for setting Rating data from Metron.

    Attributes:
        id (int): The Metron identification number for the rating.
        name (str): The name of the rating.
    """

    def __init__(self, **kwargs):
        """Initialize a Rating."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class RatingSchema(Schema):
    """Schema for Ratings."""

    id = fields.Int()
    name = fields.Str()

    class Meta:
        """Any unknown fields will be excluded."""

        unknown = EXCLUDE

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the rating object.

        Args:
            data (Any): Data from the Metron response
            **kwargs (Any): Any additional keyword arguments.

        Returns:
            A :obj: `Rating` object.
        """
        return Rating(**data)



"""
Issue module.

This module provides the following classes:

- Role
- RolesSchema
- RoleList
- Credit
- CreditsSchema
- Issue
- IssueSchema
- IssuesList
"""
from marshmallow import EXCLUDE, Schema, ValidationError, fields, post_load

from mokkari import exceptions
from mokkari.arc import ArcSchema
from mokkari.character import CharacterSchema
from mokkari.publisher import PublisherSchema
from mokkari.rating import RatingSchema
from mokkari.reprint import ReprintSchema
from mokkari.series import SeriesSchema
from mokkari.team import TeamSchema
from mokkari.variant import VariantSchema


class Role:
    """
    The Role object contains information for creators' role.

    Args:
        **kwargs (Any): The keyword arguments is used for setting role data from Metron.

    Attributes:
        id (int): The Metron identification number for the role.
        name (str): The name of the role.
    """

    def __init__(self, **kwargs):
        """Initialize a new Role."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class RolesSchema(Schema):
    """Schema for the Roles."""

    id = fields.Int()
    name = fields.Str()

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the Role object.

        Args:
            data (Any): Data from Metron response.
            **kwargs (Any): Any additional keyword arguments.

        Returns:
            A :obj:`Role` object.
        """
        return Role(**data)


class RoleList:
    """
    The :obj:`RoleList` object contains a list of roles.

    Attributes:
        id (int): The Metron identification number for the role.
        name (str): The name of the role.

    Returns:
        A list of roles.
    """

    def __init__(self, response):
        """Initialize a new RoleList."""
        self.roles = []

        schema = RolesSchema()
        for role_dict in response["results"]:
            try:
                result = schema.load(role_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error) from error

            self.roles.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.roles)

    def __len__(self):
        """Return the length of the object."""
        return len(self.roles)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.roles[index]


class Credit:
    """
    The Credit object contains information for creators credits for an issue.

    Args:
        **kwargs (Any): The keyword arguments is used for setting credits data from Metron.

    Attributes:
        id (int): The Metron identification number for the credit.
        creator (str): The name of the creator.
        role (RoleList): A list of roles for the creator.
    """

    def __init__(self, **kwargs):
        """Initialize a new Credit."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class CreditsSchema(Schema):
    """Schema for the Credits."""

    id = fields.Int()
    creator = fields.Str()
    role = fields.Nested(RolesSchema, many=True)

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the Credit object.

        Args:
            data (Any): Data from Metron response.
            **kwargs (Any): Any additional keyword arguments.

        Returns:
            A :obj:`Credit` object.
        """
        return Credit(**data)


class Issue:
    """
    The Issue object contains information for an issue.

    Args:
        **kwargs (Any): The keyword arguments is used for setting creator data from Metron.

    Attributes:
        id (int): The Metron identification number for the creator.
        publisher (Publisher): The publisher information for the issue.
        series (Series): The series information for the issue.
        number (str): The issue number.
        collection_title (str): The title of a Trade Paperback.
        story_titles (list[str]): A list of stories contained in the issue.
        cover_date (date): The cover date of the issue.
        store_date (date, optional): The date the issue went for sale.
        price (decimal): The price of the issue.
        rating (Rating): The issue rating.
        sku (str): Stock keeping unit for the issue.
        upc (str): UPC barcode for the issue.
        page_count (int): Number of pages for the issue.
        desc (str): Summary description for the issue.
        image (url): The url for a cover image associated with the issue.
        arcs (list[:obj:`Arc`]): A list of story arcs.
        credits (list[:obj:`Credit`]): A list of creator credits for the issue.
        characters (list[:obj:`Character`]): A list of characters who appear in the issue.
        teams (list[:obj:`Team`]): A list of teams who appear in the issue.
        reprints (list[:obj:`Reprint`]): A list of reprinted issue contained in the issue.
        issue_name (str): The name used to identified the issue.
        variants (list[:obj:`Variant`]): A list of variant covers for the issue.
        cv_id (int): Comic Vine ID for the issue.
        resource_url (url): The url for the resource.
        modified (datetime): The date/time the issue was last changed.
    """

    def __init__(self, **kwargs):
        """Initialize a new Issue."""
        for k, v in kwargs.items():
            setattr(self, k, v)


class IssueSchema(Schema):
    """
    Schema for the Issue API.

    .. versionchanged:: 0.1.6

        - ``name`` field  changed to ``story_titles``
        - ``__str__`` field change to ``issue_name``

    .. versionchanged:: 0.2.0
        Added ``price`` and ``sku`` fields

    .. versionchanged:: 0.2.2
        Added ``upc`` field

    .. versionchanged:: 0.2.4

        - Added ``page_count`` field
        - Changed ``price`` field from a string to float value.

    .. versionchanged:: 1.0.0

        - Changed ``price`` field to a decimal type.
        - Added ``modified`` field

    .. versionchanged:: 2.1.0

        - Add ``reprints`` field

    .. versionadded:: 2.2.2

        - Add ``collection_title`` field

    .. versionchanged:: 2.3.0

        - Removed ``volume`` field. The series object will have that information.

    .. versionchanged:: 2.3.2

        - Added ``rating`` field.

    .. versionchanged:: 2.3.3

        - Added ``resource_url`` field.

    .. versionadded:: 2.4.0

        - Added ``cv_id`` field.
    """

    id = fields.Int()
    publisher = fields.Nested(PublisherSchema)
    series = fields.Nested(SeriesSchema)
    number = fields.Str()
    collection_title = fields.Str(allow_none=True, data_key="title")
    story_titles = fields.List(fields.Str(allow_none=True), data_key="name")
    cover_date = fields.Date()
    store_date = fields.Date(allow_none=True)
    price = fields.Decimal(places=2, allow_none=True)
    rating = fields.Nested(RatingSchema)
    sku = fields.Str()
    upc = fields.Str()
    page_count = fields.Int(allow_none=True, data_key="page")
    desc = fields.Str(allow_none=True)
    image = fields.URL(allow_none=True)
    arcs = fields.Nested(ArcSchema, many=True)
    credits = fields.Nested(CreditsSchema, many=True)
    characters = fields.Nested(CharacterSchema, many=True)
    teams = fields.Nested(TeamSchema, many=True)
    reprints = fields.Nested(ReprintSchema, many=True)
    issue_name = fields.Str(data_key="issue")
    variants = fields.Nested(VariantSchema, many=True)
    cv_id = fields.Int(allow_none=True)
    resource_url = fields.URL()
    modified = fields.DateTime()

    class Meta:
        """Any unknown fields will be excluded."""

        unknown = EXCLUDE
        datetime = "%Y-%m-%dT%H:%M:%S%z"

    @post_load
    def make_object(self, data, **kwargs):
        """
        Make the issue object.

        Args:
            data (Any): Data from Metron response.
            **kwargs (Any): Any additional keyword arguments.

        Returns:
            An :obj:`Issue` object.
        """
        return Issue(**data)


class IssuesList:
    """
    The :obj:`IssuesList` object contains a list of issues.

    Attributes:
        id (int): The Metron identification number for the issue.
        issue (str): The name of the issue.
        cover_date (date): The cover date for the issue.
        modified (datetime): The date/time the creator was last changed.

    Returns:
        A list of issues.
    """

    def __init__(self, response):
        """Initialize a new IssuesList."""
        self.issues = []

        schema = IssueSchema()
        for issue_dict in response["results"]:
            try:
                result = schema.load(issue_dict)
            except ValidationError as error:
                raise exceptions.ApiError(error) from error

            self.issues.append(result)

    def __iter__(self):
        """Return an iterator object."""
        return iter(self.issues)

    def __len__(self):
        """Return the length of the object."""
        return len(self.issues)

    def __getitem__(self, index: int):
        """Return the object of a at index."""
        return self.issues[index]
